/*
 * Jalisto - JAva LIght STOrage
 * Copyright (C) 2000-2005 Xcalia http://www.xcalia.com
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Xcalia
 * 71, rue Desnouettes
 * 75014 Paris - France
 * http://www.xcalia.com
 */
package org.objectweb.jalisto.se.impl.lock;

import org.objectweb.jalisto.se.api.internal.DataWrapper;
import org.objectweb.jalisto.se.api.internal.multi.SessionMulti;
import org.objectweb.jalisto.se.exception.multi.LockException;
import org.objectweb.jalisto.se.exception.multi.SleepingException;
import org.objectweb.jalisto.se.JalistoFactory;

import java.io.Serializable;
import java.util.*;

public class Lock implements Serializable {

    private Lock() {
        currentLock = NO_LOCK;
        writeSession = "";
        transactionnalStates = new HashMap();
        readSessions = new LinkedList();
        sleepingSessions = new LinkedList();
    }

    public void release() {
        instances.add(this);
    }

    public synchronized int getCurrentAction(Object sessionId) {
        int result = ObjectTransactionnalState.NO_ACTION;
        if (transactionnalStates.containsKey(sessionId)) {
            result = ((ObjectTransactionnalState) transactionnalStates.get(sessionId)).getCurrentAction();
        }
        return result;
    }

    public synchronized DataWrapper getModifiedValue(Object sessionId) {
        ObjectTransactionnalState image = (ObjectTransactionnalState) transactionnalStates.get(sessionId);
        if (image != null) {
            return image.getCurrentValue();
        }
        return null;
    }

    public synchronized short getUpdate(Object sessionId) {
        ObjectTransactionnalState image = (ObjectTransactionnalState) transactionnalStates.get(sessionId);
        if (image != null) {
            return image.getUpdate();
        }
        return (short) -1;
    }

    public void update(Object sessionId) {
        ((ObjectTransactionnalState) transactionnalStates.get(sessionId)).update();
    }


    public synchronized void addTransactionnalValue(Object sessionId, DataWrapper value,
                                                    int action, short update) {
        // a read value can not be put over a write value
        ObjectTransactionnalState image = (ObjectTransactionnalState) transactionnalStates.get(sessionId);
        if (image == null) {
            image = new ObjectTransactionnalState(action, value, update);
            transactionnalStates.put(sessionId, image);
        } else {
            image.setValue(action, value);
            image.setUpdate(update);
        }
    }

    public synchronized void updateTransactionnalValue(Object sessionId, DataWrapper value, short update) {
        // a read value can not be put over a write value
        ObjectTransactionnalState image = (ObjectTransactionnalState) transactionnalStates.get(sessionId);
        if (image != null) {
            image.setValue(image.getCurrentAction(), value);
            image.setUpdate(update);
        }
    }

    public synchronized boolean addSleepingSession(Object sessionId) {
        if (currentLock == NO_LOCK) {
            return false;
        }
        if (sleepingSessions.contains(sessionId)) {             // TODO : cette exception n'est pas gre
            throw new LockException("dead lock : sessionId = " + String.valueOf(sessionId));
        }
        DeadLockToken asleepingSession = DeadLockToken.getToken(sessionId);
        if (sleepingSessions.isEmpty()) {
            if (currentLock == READ_LOCK) {
                for (int i = 0; i < readSessions.size(); i++) {
                    asleepingSession.waitFor(DeadLockToken.getToken(readSessions.get(i)));
                }
            } else if (currentLock == WRITE_LOCK) {
                asleepingSession.waitFor(DeadLockToken.getToken(writeSession));
            }
        } else {
            asleepingSession.waitFor(DeadLockToken.getToken(sleepingSessions.getLast()));
        }
        sleepingSessions.addLast(sessionId);
        return true;
    }

    private void wakeUpASession(Object releasingSessionId) {
        if (!sleepingSessions.isEmpty()) {
            Object awakingId = sleepingSessions.getFirst();
            DeadLockToken releasingSessionToken = DeadLockToken.getToken(releasingSessionId);
            DeadLockToken awakingSessionToken = DeadLockToken.getToken(awakingId);
            synchronized (awakingSessionToken) {
                awakingSessionToken.noMoreWaitFor(releasingSessionToken);
                if (awakingSessionToken.couldWakeUp()) {
                    sleepingSessions.remove(awakingId);
                    ((SessionMulti) JalistoFactory.getInternalFactory().getSessionById(awakingId)).wakeUp();
                }
            }
        }
    }

    public synchronized void resolveTimeOut(Object sessionId) {
        DeadLockToken timeOutToken = DeadLockToken.getToken(sessionId);
        if (sleepingSessions.contains(sessionId)) {      // wake other sessions waiting for sessionId
            int index = sleepingSessions.indexOf(sessionId);
            if ((index + 1) < sleepingSessions.size()) {
                DeadLockToken tokenWhoWaitForTimeOutToken =
                        DeadLockToken.getToken(sleepingSessions.get(index + 1));
                timeOutToken.timeOutWakeUp(tokenWhoWaitForTimeOutToken);
            } else {
                timeOutToken.timeOutWakeUp(null);
            }
            sleepingSessions.remove(sessionId);
        } else {
            // dans quels cas ?????
            DeadLockToken tokenWhoWaitForTimeOutToken = null;
            try {
                tokenWhoWaitForTimeOutToken = DeadLockToken.getToken(sleepingSessions.getFirst());
            } catch (NoSuchElementException e) {
            }
            timeOutToken.timeOutWakeUp(tokenWhoWaitForTimeOutToken);
        }
    }

    /**
     * *******************************   LOCK MANAGEMENT    *****************************************
     */

    public synchronized void takeReadLock(Object sessionId) {
        // 1 : cas pas de lock et pas de sleeping -> ajouter RL avec sessionId
        // 2 : cas pas de lock et sleepings et le premier sleeping attends la session courante -> add RL
        // 3 : cas pas de lock et sleepings et le premier sleeping n'attends pas la session courante -> sleep
        if (currentLock == NO_LOCK) {
            if (sleepingSessions.isEmpty()) { // 1
                currentLock = READ_LOCK;
            } else {
                DeadLockToken awakingToken = DeadLockToken.getToken(sleepingSessions.getFirst());
                if (awakingToken.isWaitingFor(DeadLockToken.getToken(sessionId))) { // 2
                    currentLock = READ_LOCK;
                } else { // 3
                    throw new SleepingException("takeReadLock case 3", this);
                }
            }
            // 4 : cas read session avec sessionId et eventuellement d'autre -> ne fe rien
            // 5 : cas read session avec autre exclusivement que sessionId et pas sleeping -> ajouter sessionId readSessions
            // 6 : cas read session avec autre exclusivement que sessionId et sleeping -> sleep
        } else if (currentLock == READ_LOCK) {
            if (!readSessions.contains(sessionId)) {
                if (!sleepingSessions.isEmpty()) {
                    throw new SleepingException("takeReadLock case 6", this);     // 6
                } // else  : 5
            }   // else : 4
            // 7 : cas write session avec sessionId	-> rien  faire
            // 8 : cas write session avec autre		-> sleep
        } else if (currentLock == WRITE_LOCK) {
            if (!writeSession.equals(sessionId)) { // 8
                throw new SleepingException("takeReadLock case 8", this);
            } // else  : 7
        }
        if (!readSessions.contains(sessionId)) {
            readSessions.addLast(sessionId);
        }
    }

    public synchronized void takeWriteLock(Object sessionId) {
        // 1 : cas pas de lock et pas sleepings -> ajouter WL avec sessionId
        // 2 : cas pas de lock et sleepings et premier sleeping attends sessionId -> sessionId pose WL
        // 3 : cas pas de lock et sleepings et premier sleeping n'attends pas sessionId -> sleep
        if (currentLock == NO_LOCK) {
            if (sleepingSessions.isEmpty()) { // 1
                currentLock = WRITE_LOCK;
            } else {
                DeadLockToken awakingToken = DeadLockToken.getToken(sleepingSessions.getFirst());
                if (awakingToken.isWaitingFor(DeadLockToken.getToken(sessionId))) { // 2
                    currentLock = WRITE_LOCK;
                } else { // 3
                    throw new SleepingException(this);
                }
            }
            // 4 : cas read session avec seulement sessionId -> transforme RL en WL
            // 5 : cas read session avec autre exclusivement que sessionId -> sleep
            // 6 : cas read session avec autre que sessionId ET sessionId -> sleep
        } else if (currentLock == READ_LOCK) {
            if (readSessions.contains(sessionId)) {
                if (readSessions.size() > 1) {     // 6
                    wakeUpASession(sessionId);
                    throw new SleepingException(this);
                } else { // 4
                    currentLock = WRITE_LOCK;
                }
            } else { // 5
                throw new SleepingException(this);
            }
            // 7 : cas WL session avec sessionId	-> ne fait rien
            // 8 : cas WL session avec autre		-> sleep
        } else { // currentLock == WL
            if (writeSession.equals(sessionId)) { // 7
                return;
            } else { // 8
                throw new SleepingException(this);
            }
        }
        writeSession = sessionId;
    }

    public synchronized void releaseReadLock(Object sessionId) {
        if (currentLock == WRITE_LOCK) {
            if (writeSession.equals(sessionId)) {
                readSessions.remove(sessionId);
            } else {
                throw new LockException();
            }
        } else {
            readSessions.remove(sessionId);
            if (readSessions.isEmpty()) {
                currentLock = NO_LOCK;
            }
            wakeUpASession(sessionId);
        }
    }

    public synchronized void releaseWriteLock(Object sessionId) {
        if ((currentLock == WRITE_LOCK) && (writeSession.equals(sessionId))) {
            readSessions.remove(sessionId);
            currentLock = NO_LOCK;
            writeSession = "";
            wakeUpASession(sessionId);
        } else {
            throw new LockException();
        }
    }

    public synchronized void releaseAllLock(Object sessionId) {
        if ((currentLock == WRITE_LOCK) && (writeSession.equals(sessionId))) {
            writeSession = "";
            currentLock = NO_LOCK;
        } else if ((currentLock == READ_LOCK) && (readSessions.contains(sessionId))) {
            currentLock = NO_LOCK;
        }
        readSessions.remove(sessionId);
        transactionnalStates.remove(sessionId);
        wakeUpASession(sessionId);
    }


    /**
     * *************************     TESTS     ******************************************
     */

    public synchronized boolean canBeDeleted() {
        return ((currentLock == NO_LOCK) &&
                transactionnalStates.isEmpty() &&
                sleepingSessions.isEmpty());
    }

    public synchronized boolean hasModifiedValueWith(Object sessionId) {
        return (transactionnalStates.get(sessionId) != null);
    }

    public synchronized boolean hasReadLockWith(Object sessionId) {
        return readSessions.contains(sessionId);
    }

    public synchronized boolean mustTakeWriteLock(Object sessionId) {
        ObjectTransactionnalState itm = (ObjectTransactionnalState) transactionnalStates.get(sessionId);
        return ((itm.getCurrentAction() != ObjectTransactionnalState.READ_ACTION) &&
                (itm.getCurrentAction() != ObjectTransactionnalState.NO_ACTION));
    }

    public synchronized boolean hasWriteLockWith(Object sessionId) {
        return writeSession.equals(sessionId);
    }

    public synchronized boolean hasToBeRemoveForExtentWith(Object sessionId) {
        if (transactionnalStates.containsKey(sessionId)) {
            ObjectTransactionnalState itm = (ObjectTransactionnalState) transactionnalStates.get(sessionId);
            return (itm.getCurrentAction() == ObjectTransactionnalState.DELETE_ACTION);
        }
        return false;
    }

    /**
     * **********************************************************************************************
     */

    private String getLockAsString(int lock) {
        if (lock == NO_LOCK) {
            return "no_lock";
        } else if (lock == READ_LOCK) {
            return "r_lock";
        } else {
            return "w_lock";
        }
    }


    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append(getLockAsString(currentLock)).append(" ").append(super.toString()).append("\n");
        sb.append("\tRead Locks : ").append(readSessions).append("\n");
        sb.append("\tWrite Lock : ").append(writeSession).append("\n");
        sb.append("\tActions : \n");
        sb.append("\tSleeping session : ").append(sleepingSessions).append("\n");
        Iterator sessions = transactionnalStates.keySet().iterator();
        if (!sessions.hasNext()) {
            sb.append("CurrentStates is empty");
        }
        while (sessions.hasNext()) {
            Object key = sessions.next();
            ObjectTransactionnalState image = (ObjectTransactionnalState) transactionnalStates.get(key);
            sb.append(" ").append(image);
        }
        sb.append("\n");
        return sb.toString();
    }


    private HashMap transactionnalStates;
    private LinkedList readSessions;
    private LinkedList sleepingSessions;

    private Object writeSession;                   // utiliser "" pour signifier null
    private int currentLock;


    static final long serialVersionUID = -7589377087694761459L;


    public static final int NO_LOCK = 0;
    public static final int READ_LOCK = 1;
    public static final int WRITE_LOCK = 2;

    private static ArrayList instances = new ArrayList();

    public static Lock newInstance() {
        if (instances.isEmpty()) {
            return new Lock();
        } else {
            return (Lock) instances.remove(0);
        }
    }
}
