/*
 * Jalisto - JAva LIght STOrage
 * Copyright (C) 2000-2005 Xcalia http://www.xcalia.com
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Xcalia
 * 71, rue Desnouettes
 * 75014 Paris - France
 * http://www.xcalia.com
 */
package org.objectweb.jalisto.se.impl.multi;

import org.objectweb.jalisto.se.api.*;
import org.objectweb.jalisto.se.api.physical.PluggablePhysicalFileAccess;
import org.objectweb.jalisto.se.api.internal.InTransactionBaseImage;
import org.objectweb.jalisto.se.api.internal.JalistoObject;
import org.objectweb.jalisto.se.exception.JalistoException;
import org.objectweb.jalisto.se.impl.InFileAddress;
import org.objectweb.jalisto.se.impl.server.InTransactionBaseImageImpl;
import org.objectweb.jalisto.se.impl.trace.Trace;

import java.util.Collection;
import java.util.Iterator;

public class InTransactionBaseImageMultiImpl extends InTransactionBaseImageImpl {

    protected InTransactionBaseImageMultiImpl(PluggablePhysicalFileAccess physicalAccess, JalistoProperties properties) {
        super(physicalAccess, properties);
        this.isWorking = false;
        trace.println(Trace.DEBUG, "Uses InTransactionBaseImageMultiImpl");
    }

    /**
     * ***************************************************************************************************
     */

    public synchronized void commit() {
        isWorking = true;
        physicalAccess.startCommit();
        try {
            Iterator values = deletedObjects.values().iterator();
            while (values.hasNext()) {
                physicalAccess.deleteFileObject((InFileAddress) values.next());
            }
            values = newObjects.values().iterator();
            while (values.hasNext()) {
                JalistoObject fo = (JalistoObject) values.next();
                physicalAccess.insertFileObject(fo.getIfa(), fo);
            }
            values = modifiedObjects.values().iterator();
            while (values.hasNext()) {
                JalistoObject fo = (JalistoObject) values.next();
                physicalAccess.updateFileObject(fo.getIfa(), fo);
            }
            finishCommit();
            physicalAccess.commit();
        } finally {
            isWorking = false;
        }
    }

    public synchronized void rollback() {
        isWorking = true;
        try {
            clear();
            readObjects.clear();
            physicalAccess.rollback();
        } finally {
            isWorking = false;
        }
    }


    public synchronized void insertFileObject(InFileAddress ifa, JalistoObject fo) {
        isWorking = true;
        try {
            String address = ifa.getAddress();
            if (modifiedObjects.containsKey(address) ||
                newObjects.containsKey(address) ||
                readObjects.containsKey(address)) {
                isWorking = false;
                throw new JalistoException("you could not insert a object already in base : " + ifa + ", " + fo);
            }
            if (deletedObjects.containsKey(address)) {
                deletedObjects.remove(address);
                modifiedObjects.put(address, fo);
            } else {
                newObjects.put(address, fo);
            }
        } finally {
            isWorking = false;
        }
    }

    public synchronized JalistoObject readFileObjectAt(InFileAddress ifa) {
        isWorking = true;
        try {
            String address = ifa.getAddress();
            if (deletedObjects.containsKey(address)) {
                throw new JalistoException("object has been removed from base : " + ifa);
            }
            JalistoObject result = (JalistoObject) readObjects.get(address);
            if (result != null)
                return result;
            result = (JalistoObject) modifiedObjects.get(address);
            if (result != null)
                return result;
            result = (JalistoObject) newObjects.get(address);
            if (result != null)
                return result;
            result = physicalAccess.readFileObjectAt(ifa);
            readObjects.put(address, result);
            return result;
        } finally {
            isWorking = false;
        }

    }

    public synchronized void updateFileObject(InFileAddress ifa, JalistoObject fo) {
        isWorking = true;
        try {
            String address = ifa.getAddress();
            if (deletedObjects.containsKey(address)) {
                isWorking = false;
                throw new JalistoException("object has been removed from base : " + ifa + ", " + fo);
            }
            if (newObjects.containsKey(address)) {
                newObjects.put(address, fo);
            } else {
                modifiedObjects.put(address, fo);
            }
            readObjects.remove(address);
        } finally {
            isWorking = false;
        }
    }

    public synchronized void deleteFileObject(InFileAddress ifa) {
        isWorking = true;
        try {
            String address = ifa.getAddress();
            if (deletedObjects.containsKey(address)) {
                isWorking = false;
                throw new JalistoException("object has already been removed from base : " + ifa);
            }
            if (newObjects.remove(address) == null) {
                modifiedObjects.remove(address);
            }
            readObjects.remove(address);
            deletedObjects.put(address, ifa);
        } finally {
            isWorking = false;
        }
    }


    public boolean isWorking() {
        return isWorking;
    }

    public synchronized boolean startWorking() {
        if (isWorking) {
            return false;
        }
        isWorking = true;
        return true;
    }

    public synchronized void stopWorking() {
        isWorking = false;
    }

    public synchronized Collection getKeysStartingWith(String filter, boolean withOrder) {
        return super.getKeysStartingWith(filter, withOrder);
    }

    /**
     * ***************************    INTERNAL    *****************************************************
     */

    protected void finishCommit() {
        if (keepInMemory) {
            readObjects.putAll(modifiedObjects);
            readObjects.putAll(newObjects);
        }
        clear();
    }

    protected void clear() {
        if (!keepInMemory) {
            readObjects.clear();
        }
        modifiedObjects.clear();
        newObjects.clear();
        deletedObjects.clear();
    }


    private boolean isWorking;


    /**
     * *******************************    STATIC CONSTRUCTOR    ******************************************
     */

    public static InTransactionBaseImage getInTransactionBaseImageInstance(
            PluggablePhysicalFileAccess physicalAccess, JalistoProperties properties) {
        return new InTransactionBaseImageMultiImpl(physicalAccess, properties);
    }
}
