/*
 * Jalisto - JAva LIght STOrage
 * Copyright (C) 2000-2005 Xcalia http://www.xcalia.com
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Xcalia
 * 71, rue Desnouettes
 * 75014 Paris - France
 * http://www.xcalia.com
 */
package org.objectweb.jalisto.se.impl.server;

import org.objectweb.jalisto.se.api.JalistoProperties;
import org.objectweb.jalisto.se.api.physical.PluggablePhysicalFileAccess;
import org.objectweb.jalisto.se.api.internal.*;
import org.objectweb.jalisto.se.exception.JalistoException;
import org.objectweb.jalisto.se.impl.trace.Trace;
import org.objectweb.jalisto.se.impl.InFileAddress;
import org.objectweb.jalisto.se.JalistoFactory;

import java.util.*;

public class InTransactionBaseImageImpl implements InTransactionBaseImage {

    protected InTransactionBaseImageImpl(PluggablePhysicalFileAccess physicalAccess, JalistoProperties properties) {
        this.trace = JalistoFactory.getInternalFactory().getTracer(properties);
        this.keepInMemory = properties.isKeepingInMemory();
        this.physicalAccess = physicalAccess;
        this.deletedObjects = new HashMap();
        this.newObjects = new HashMap();
        this.modifiedObjects = new HashMap();
        this.readObjects = JalistoFactory.getInternalFactory().getCache(properties, properties.getPageCacheSize(),
                                                          "inTransactionBaseImageCache");
    }


    public void commit() {
        physicalAccess.startCommit();
        Iterator values = deletedObjects.values().iterator();
        while (values.hasNext()) {
            physicalAccess.deleteFileObject((InFileAddress) values.next());
        }
        values = newObjects.values().iterator();
        while (values.hasNext()) {
            JalistoObject fo = (JalistoObject) values.next();
            physicalAccess.insertFileObject(fo.getIfa(), fo);
        }
        values = modifiedObjects.values().iterator();
        while (values.hasNext()) {
            JalistoObject fo = (JalistoObject) values.next();
            physicalAccess.updateFileObject(fo.getIfa(), fo);
        }
        finishCommit();
        physicalAccess.commit();
    }

    public void rollback() {
        clear();
        readObjects.clear();
        physicalAccess.rollback();
    }

    protected void finishCommit() {
        if (keepInMemory) {
            readObjects.putAll(modifiedObjects);
            readObjects.putAll(newObjects);
        }
        clear();
    }

    protected void clear() {
        if (!keepInMemory) {
            readObjects.clear();
        }
        modifiedObjects.clear();
        newObjects.clear();
        deletedObjects.clear();
    }


    public void insertFileObject(InFileAddress ifa, JalistoObject fo) {
        String address = ifa.getAddress();
        if (modifiedObjects.containsKey(address) ||
            newObjects.containsKey(address) ||
            readObjects.containsKey(address)) {
            throw new JalistoException("you could not insert a object already in base : " + ifa + ", " + fo);
        }
        if (deletedObjects.containsKey(address)) {
            deletedObjects.remove(address);
            modifiedObjects.put(address, fo);
        } else {
            newObjects.put(address, fo);
        }
    }

    public JalistoObject readFileObjectAt(InFileAddress ifa) {
        String address = ifa.getAddress();
        if (deletedObjects.containsKey(address)) {
            throw new JalistoException("object has been removed from base : " + ifa);
        }
        JalistoObject result = (JalistoObject) readObjects.get(address);
        if (result != null)
            return result;
        result = (JalistoObject) modifiedObjects.get(address);
        if (result != null)
            return result;
        result = (JalistoObject) newObjects.get(address);
        if (result != null)
            return result;
        result = physicalAccess.readFileObjectAt(ifa);
        readObjects.put(address, result);
        return result;
    }

    public void updateFileObject(InFileAddress ifa, JalistoObject fo) {
        String address = ifa.getAddress();
        if (deletedObjects.containsKey(address)) {
            throw new JalistoException("object has been removed from base : " + ifa + ", " + fo);
        }
        if (newObjects.containsKey(address)) {
            newObjects.put(address, fo);
        } else {
            modifiedObjects.put(address, fo);
        }
        readObjects.remove(address);
    }

    public void deleteFileObject(InFileAddress ifa) {
        String address = ifa.getAddress();
        if (deletedObjects.containsKey(address)) {
            throw new JalistoException("object has already been removed from base : " + ifa);
        }
        if (newObjects.remove(address) == null) {
            modifiedObjects.remove(address);
        }
        readObjects.remove(address);
        deletedObjects.put(address, ifa);
    }


    public boolean isWorking() {
        return false;
    }

    public boolean startWorking() {
        return true;
    }

    public void stopWorking() {
    }


    public int getDeletedObjectsSize() {
        return deletedObjects.size();
    }

    public int getModifiedObjectsSize() {
        return modifiedObjects.size();
    }

    public int getNewObjectsSize() {
        return newObjects.size();
    }

    public Map getReadObjects() {
        return readObjects;
    }

    public boolean isKeepInMemory() {
        return keepInMemory;
    }

    public void setKeepInMemory(boolean keepInMemory) {
        this.keepInMemory = keepInMemory;
    }

    public Collection getKeysStartingWith(String filter, boolean withOrder) {
        ArrayList keys = new ArrayList(physicalAccess.getKeysStartingWith(filter));
        Iterator newKeys = newObjects.keySet().iterator();
        while (newKeys.hasNext()) {
            String key = (String) newKeys.next();
            if (key.startsWith(filter)) {
                keys.add(key);
            }
        }
        Iterator deletedKeys = deletedObjects.keySet().iterator();
        while (deletedKeys.hasNext()) {
            String key = (String) deletedKeys.next();
            if (key.startsWith(filter)) {
                keys.remove(key);
            }
        }
        if (withOrder) {
            Collections.sort(keys);
        }
        return keys;
    }


    protected Map deletedObjects;
    protected Map modifiedObjects;
    protected Map newObjects;
    protected Map readObjects;
    protected PluggablePhysicalFileAccess physicalAccess;
    protected Trace trace;
    protected boolean keepInMemory;


    /**
     * *******************************    STATIC CONSTRUCTOR    ******************************************
     */

    public static InTransactionBaseImage getInTransactionBaseImageInstance(
            PluggablePhysicalFileAccess physicalAccess, JalistoProperties properties) {
        return new InTransactionBaseImageImpl(physicalAccess, properties);
    }
}
