/*
 * Jalisto - JAva LIght STOrage
 * Copyright (C) 2000-2005 Xcalia http://www.xcalia.com
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Xcalia
 * 71, rue Desnouettes
 * 75014 Paris - France
 * http://www.xcalia.com
 */
package org.objectweb.jalisto.se.impl.server;

import org.objectweb.jalisto.se.api.*;
import org.objectweb.jalisto.se.api.internal.*;
import org.objectweb.jalisto.se.api.internal.multi.LockManager;
import org.objectweb.jalisto.se.api.internal.InternalPhysicalFileAccess;
import org.objectweb.jalisto.se.api.query.QueryManager;
import org.objectweb.jalisto.se.exception.*;
import org.objectweb.jalisto.se.impl.trace.Trace;
import org.objectweb.jalisto.se.impl.InFileAddress;
import org.objectweb.jalisto.se.impl.LogicalOid;
import org.objectweb.jalisto.se.impl.ExtentImpl;
import org.objectweb.jalisto.se.impl.TransactionImpl;
import org.objectweb.jalisto.se.JalistoFactory;

import java.io.Serializable;
import java.util.*;

public abstract class SessionImpl implements SessionInternal {

    public SessionImpl(JalistoProperties properties) {
        InternalFactory internalFactory = JalistoFactory.getInternalFactory();
        trace = internalFactory.getTracer(properties);
        sessionId = internalFactory.getSessionId(properties, this);
        trace.println(Trace.SESSION, "{0} : create new instance with properties : {1}", sessionId, properties);
        cache = internalFactory.getCache(properties, properties.getObjectCacheSize(), "object cache");
        fileAccess = internalFactory.getLogicalAccess(properties);
        oidTable = internalFactory.getOidTable(properties);
        repository = internalFactory.getMetaRepository(properties);
        props = properties;
        transaction = createTransaction();
        checkStateInBase();
        trace.println(Trace.DEBUG, "properties for session {0} : {1}", sessionId, properties);
    }

    protected void checkSerialisable(Object[] object) {
        if (object != null) {
            for (int i = 0; i < object.length; i++) {
                Object o = object[i];
                if ((o != null) && !(o instanceof Serializable)) {
                    throw new JalistoException("not serializabel class in data : "+o.getClass());
                }
            }
        }
    }


    public void defineClass(ClassDescription classMetaDescription) {
        trace.println(Trace.SESSION, "{0} : defineClass({1})", sessionId, classMetaDescription);
        checkIsOpen("defineClass");
        repository.defineClass(this, classMetaDescription);
    }

    public void removeClass(String fullClassName) {
        trace.println(Trace.SESSION, "{0} : removeClass({1})", sessionId, fullClassName);
        checkIsOpen("removeClass");
        repository.removeClass(this, fullClassName);
    }

    public Collection getAllClassNames() {
        return repository.getAllClassNames();
    }

    public boolean isClassDefined(String fullQualifiedClassName) {
        checkIsOpen("isClassDefined");
        try {
            repository.getClidFromClassName(fullQualifiedClassName);
            return true;
        } catch (SchemaException e) {
        }
        return false;
    }

    public String getClassNameFor(Object oid) {
        trace.println(Trace.SESSION, "{0} : getClassNameFor({1})", sessionId, oid);
        checkIsOpen("getClassNameFor");
        LogicalOid floid = getFloidFromOid(oid);
        return repository.getClassNameFromClid(new Short(floid.getClid()));
    }

    public Object createObject(Object[] objectToCreate, Class objectClass) {
        checkSerialisable(objectToCreate);
        return createObject(makeNewFileOid(objectClass), objectToCreate);
    }

    public Object createObject(Object[] objectToCreate, String objectClassName) {
        checkSerialisable(objectToCreate);
        return createObject(makeNewFileOid(objectClassName), objectToCreate);
    }

    public Object[] readObjectByOid(Object oid) {
        return readObjectByOid(oid, true);
    }

    public Collection readObjectsByOids(Collection oids) {
        Iterator oidsIterator = oids.iterator();
        ArrayList result = new ArrayList();
        while(oidsIterator.hasNext()) {
            result.add(readObjectByOid(oidsIterator.next()));
        }
        return result;
    }

    public boolean contains(Object oid) {
        trace.println(Trace.SESSION, "{0} contains({1})", sessionId, oid);
        checkValidity("contains", true);
        return oidTable.containsFloid(sessionId, (LogicalOid) oid);
    }

    public Extent getExtent(Class theClass) {
        String fullClassName = theClass.getName();
        return getExtent(fullClassName);
    }

    public Extent getExtent(String fullClassName) {
        trace.println(Trace.SESSION, "{0} getExtent({1})", sessionId, fullClassName);
        checkValidity("getExtent", true);
        return new ExtentImpl(repository.getClidFromClassName(fullClassName), this);
    }

    public Collection getFloidsFromClid(Object sessionId, Object clid) {
        return oidTable.getFloidsFromClid(sessionId, clid, false);
    }

    public Long reserveFloids(short clid, int number) {
        return new Long(oidTable.getIdentityProvider().reserveFloids(clid, number));
    }

    public void setOptimistic() {
    }

    public void setPessimistic() {
    }

    public void begin() {
        trace.println(Trace.SESSION, "{0} : begin()", sessionId);
        checkIsOpen("begin");
        oidTable.begin(sessionId);
        fileAccess.begin(sessionId);
    }

    public void commit() {
        trace.println(Trace.SESSION, "{0} : commit()", sessionId);
        checkIsOpen("commit");
        oidTable.commit(sessionId);
        fileAccess.commit(sessionId);
    }

    public void rollback() {
        trace.println(Trace.SESSION, "{0} : rollback()", sessionId);
        checkIsOpen("rollback");
        oidTable.rollback(sessionId);
        fileAccess.rollback(sessionId);
    }

    public void openSession() {
        trace.println(Trace.SESSION, "{0} : openSession()", sessionId);
        if (isOpen) {
            throw new TransactionException("Session is already open");
        }
        oidTable.open(sessionId);
        fileAccess.getPhysicalAccess().open();
        repository.addSession(this);
        isOpen = true;
    }

    public void closeSession() {
        trace.println(Trace.SESSION, "{0} : closeSession()", sessionId);
        if (!isOpen) {
            throw new TransactionException("Session is not open");
        }
        oidTable.close(sessionId);
        fileAccess.getPhysicalAccess().close();
        repository.removeSession(this);
        isOpen = false;
    }

    public boolean isOpen() {
        return isOpen;
    }

    public void checkValidity(String message, boolean mustBeActive) {
        checkIsOpen(message);
        if (mustBeActive) {
            if (!transaction.isActive()) {
                throw new TransactionException("Transaction must be active during " + message);
            }
        } else {
            if (transaction.isActive()) {
                throw new TransactionException("Transaction must not be active during " + message);
            }
        }
    }

    public boolean isRemoteSession() {
        return false;
    }

    public QueryManager getQueryManager() {
        checkIsOpen("getQueryManager");
        if (queryManager == null) {
            try {
                String queryManagerClassName = props.getProperty(JalistoProperties.QUERY_MANAGER_CLASS_KEY);
                queryManager = (QueryManager) Class.forName(queryManagerClassName).newInstance();
                queryManager.init(this);
            } catch (InstantiationException e) {
                trace.println(Trace.TRACE_ON, "cannot instanciate query manager : {0}", e.getMessage());
            } catch (IllegalAccessException e) {
                trace.println(Trace.TRACE_ON, "cannot initialize query manager : {0}", e.getMessage());
            } catch (ClassNotFoundException e) {
                trace.println(Trace.TRACE_ON, "cannot  find query manager class : {0}", e.getMessage());
            }
        }
        return queryManager;
    }

    public Transaction currentTransaction() {
        return transaction;
    }

    public LogicalSystemPageAccess getFileAccess() {
        return fileAccess;
    }

    public MetaRepository getMetaRepository() {
        checkIsOpen("getMetaRepository");
        return repository;
    }

    public SessionInternal getInternalSession() {
        return this;
    }

    public JalistoProperties getProperties() {
        return props;
    }

    public OidTable getOidTable() {
        return oidTable;
    }

    public LockManager getLockManager() {
        return null;
    }

    public boolean isNewBase() {
        return fileAccess.getPhysicalAccess().isNewBase();
    }

    public Object getSessionId() {
        return sessionId;
    }

    public void clearObjectCache() {
        trace.println(Trace.SESSION, "{0} : clearObjectCache()", sessionId);
        checkIsOpen("clearObjectCache");
        cache.clear();
    }

    public Object[] refreshObjectByOid(Object oid) {
        return readObjectByOid(oid);
    }

    public void reorganize() {
        trace.println(Trace.SESSION, "{0} : reorganize()", sessionId);
        if (!props.allowsSpecialFunctionnalities()) {
            throw new TransactionException("The reorganize operation is not allowed");
        }
        if (isOpen) {
            throw new TransactionException("Session must not be open during reorganize()");
        }
        if (transaction.isActive()) {
            throw new JalistoException("Transaction must not be active during reorganize()");
        }
        repository.checkNoSessions(
                "There are active sessions during attempt to reorganize datastore " + getSessionId());
        try {
            openSession();
            transaction.begin();
            fileAccess.cleanAllNullValue();
            transaction.commit();
            fileAccess.getPhysicalAccess().getInternalAccess().reorganize();
            closeSession();
        } catch (Exception e) {
            throw new JalistoException(e);
        }
    }

    public void eraseStorage() {
        trace.println(Trace.SESSION, "{0} : eraseStorage()", sessionId);
        if (!props.allowsSpecialFunctionnalities()) {
            throw new TransactionException("The eraseStorage operation is not allowed");
        }
        if (isOpen) {
            throw new TransactionException("Session must not be open during eraseStorage()");
        }
        if (transaction.isActive()) {
            throw new JalistoException("Transaction must not be active during eraseStorage()");
        }
        repository.checkNoSessions(
                "There are active sessions during attempt to eraseStorage datastore " + getSessionId());

        try {
            openSession();
            Iterator classes = (new ArrayList(repository.getAllClassNames())).iterator();
            while (classes.hasNext()) {
                String className = (String) classes.next();
                removeClass(className);
            }
            transaction.begin();
            oidTable.getIdentityProvider().resetAll();
            transaction.commit();
            closeSession();
        } catch (Exception e) {
            throw new JalistoException(e);
        }
    }

    public String toString() {
        return "session " + sessionId;
    }

    /**
     * *************************************    PRIVATE    *************************************
     */

    protected void checkIsOpen(String message) {
        if (!isOpen) {
            throw new TransactionException("Session must be open during " + message);
        }
    }

    protected LogicalOid getFloidFromOid(Object oid) {
        try {
            return (LogicalOid) oid;
        } catch (ClassCastException cce) {
            throw new IdentityException("oid is not a FileLoid but a " + oid.getClass().getName());
        }
    }

    private void checkStateInBase() {
        trace.println(Trace.SESSION, "{0} : checkStateInBase()", sessionId);
        JalistoProperties state;
        InternalPhysicalFileAccess physical = fileAccess.getPhysicalAccess();

        try {
            state = (JalistoProperties) physical.readFileObjectAt(BSTATE_IFA).getData();
        } catch (JalistoException e) {
            physical.writeObjectInBase(BSTATE_IFA, props.getClone(), false);
            return;
        }
        props.compareProperties(state);
    }

    protected Transaction createTransaction() {
        return new TransactionImpl(this);
    }


    protected Map cache;
    protected OidTable oidTable;
    protected LogicalSystemPageAccess fileAccess;
    protected InternalMetaRepository repository;
    protected QueryManager queryManager;

    protected Transaction transaction;
    protected boolean isOpen;

    protected JalistoProperties props;
    protected Object sessionId;

    protected Trace trace;

    public static final InFileAddress BSTATE_IFA = new InFileAddress("bState");
}
