/*
 * Jalisto - JAva LIght STOrage
 * Copyright (C) 2000-2005 Xcalia http://www.xcalia.com
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Xcalia
 * 71, rue Desnouettes
 * 75014 Paris - France
 * http://www.xcalia.com
 */
package org.objectweb.jalisto.se.storage.raf;

import org.objectweb.jalisto.se.impl.trace.Trace;
import org.objectweb.jalisto.se.impl.util.JalistoUtils;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.*;

public class RecordsFile {
    // Total length in bytes of the global database headers.
    protected static final int FILE_HEADERS_REGION_LENGTH = 16;

    // Number of bytes in the record header.
    protected static final int RECORD_HEADER_LENGTH = 16;

    // File pointer to the num records header.
    protected static final long NUM_RECORDS_HEADER_LOCATION = 0;

    // File pointer to the data start pointer header.
    protected static final long DATA_START_HEADER_LOCATION = 4;

    protected double additionalSpace;

    // The total length of one index entry - the key length plus the record header length.
    protected int indexEntryLength;

    // The length of a key in the index. default = 64
    protected int maxKeyLength;

    // Current file pointer to the start of the record data.
    protected long dataStartPtr;

    protected Trace trace;

    // The underlying file.
    protected RandomAccessFile file;

    /**
     * Hashtable which holds the in-memory index. For efficiency, the entire index
     * is cached in memory. The hashtable maps a key of type String to a RecordHeader.
     */
    protected HashMap memIndex;


    /**
     * Creates a new database file.  The initialSize parameter determines the
     * amount of space which is allocated for the index.  The index can grow
     * dynamically, but the parameter is provide to increase
     * efficiency.
     */
    public RecordsFile(String dbPath)
            throws IOException, RecordsFileException {
        File f = new File(dbPath);
        if (f.exists()) {
            throw new RecordsFileException("Database already exits: " + dbPath);
        }
        file = new RandomAccessFile(f, "rw");
    }

    /**
     * Opens an existing database file and initializes the dataStartPtr. The accessFlags
     * parameter can be "r" or "rw" -- as defined in RandomAccessFile.
     * Initialize the in-memory index.
     */
    public RecordsFile(String dbPath, String accessFlags)
            throws IOException, RecordsFileException {
        File f = new File(dbPath);
        if (!f.exists()) {
            throw new RecordsFileException("Database not found: " + dbPath);
        }
        file = new RandomAccessFile(f, accessFlags);
    }

    public void initNewStorage(int initialSize, int maxKeyLength, int additionalSpace)
            throws IOException, RecordsFileException {
        this.maxKeyLength = maxKeyLength;   //   to have string size == maxKeyLength
        this.additionalSpace = (((double) additionalSpace) / 100) + 1;
        this.indexEntryLength = maxKeyLength + RECORD_HEADER_LENGTH;
        this.dataStartPtr = indexPositionToKeyFp(initialSize); // Record Data Region starts were the (i+1)th index entry would start.
        setFileLength(dataStartPtr);
        writeNumRecordsHeader(0);
        writeDataStartPtrHeader(dataStartPtr);
        memIndex = new HashMap(initialSize);
    }

    public void initExistingStorage(int maxKeyLength, int additionalSpace)
            throws IOException, RecordsFileException {
        this.maxKeyLength = maxKeyLength;      //   to have string size == maxKeyLength
        this.additionalSpace = (((double) additionalSpace) / 100) + 1;
        this.indexEntryLength = maxKeyLength + RECORD_HEADER_LENGTH;
        this.dataStartPtr = readDataStartHeader();
        readHeadersInMemIndex();
    }

    public synchronized void close() throws IOException, RecordsFileException {
        try {
            file.close();
        } finally {
            memIndex.clear();
            memIndex = null;
            file = null;
        }
    }

    /**
     * *********************************************************************************************
     */

    public synchronized void insertRecord(RecordWriter rw) throws RecordsFileException, IOException {
        String key = rw.getKey();
        if (recordExists(key)) {
            throw new RecordsFileException("Key exists: " + key);
        }
        insureIndexSpace(getNumRecords() + 1);
        int length = (new Double(rw.getDataLength() * additionalSpace)).intValue();
        RecordHeader newRecord = allocateRecord(key, length);
        writeRecordData(newRecord, rw);
        addEntryToIndex(key, newRecord, getNumRecords());
    }

    public synchronized RecordReader readRecord(String key) throws RecordsFileException, IOException {
        byte[] data = readRecordData(key);
        return new RecordReader(key, data);
    }

    /**
     * Updates an existing record. If the new contents do not fit in the original record,
     * then the update is handled by deleting the old record and adding the new.
     */
    public synchronized void updateRecord(RecordWriter rw)
            throws RecordsFileException, IOException {
        RecordHeader header = keyToRecordHeader(rw.getKey());

        if (rw.getDataLength() > header.getDataCapacity()) {
            deleteRecord(rw.getKey());
            insertRecord(rw);
        } else {
            writeRecordData(header, rw);
            writeRecordHeaderToIndex(header);
        }
    }

    public synchronized void deleteRecord(String key) throws RecordsFileException, IOException {
        RecordHeader delRec = keyToRecordHeader(key);
        int currentNumRecords = getNumRecords();

        if (getFileLength() == (delRec.getDataPointer() + delRec.getDataCapacity())) {
            // shrink file since this is the last record in the file
            setFileLength(delRec.getDataPointer());
        } else {
            RecordHeader previous = getRecordAt(delRec.getDataPointer() - 1);

            if (previous != null) {
                // append space of deleted record onto previous record
                previous.setDataCapacity(previous.getDataCapacity() + delRec.getDataCapacity());
                writeRecordHeaderToIndex(previous);
            } else {
                // target record is first in the file and is deleted by adding its space to
                // the second record.
                RecordHeader secondRecord = getRecordAt(delRec.getDataPointer() + (long) delRec.getDataCapacity());
                byte[] data = readRecordData(secondRecord);
                secondRecord.setDataPointer(delRec.getDataPointer());
                secondRecord.setDataCapacity(secondRecord.getDataCapacity() + delRec.getDataCapacity());
                writeRecordData(secondRecord, data);
                writeRecordHeaderToIndex(secondRecord);
            }
        }
        deleteEntryFromIndex(key, delRec, currentNumRecords);
    }


    protected void setFileLength(long l) throws IOException {
        file.setLength(l);
    }

    protected long getFileLength() throws IOException {
        return file.length();
    }

    /**
     * Appends an entry to end of index. Assumes that insureIndexSpace() has already been called.
     */
    protected void addEntryToIndex(String key, RecordHeader newRecord, int currentNumRecords)
            throws IOException, RecordsFileException {
        DbByteArrayOutputStream temp = DbByteArrayOutputStream.getInstance();
        temp.getOutput().writeUTF(key);

        if (temp.size() > maxKeyLength) {
            throw new RecordsFileException("Key " + key + " of size " + temp.size() +
                                           " is larger than permitted size of " +
                                           maxKeyLength + " bytes");
        }

        writeOutputStream(indexPositionToKeyFp(currentNumRecords), temp);
        temp.reset();
        writeRecordHeader(indexPositionToRecordHeaderFp(currentNumRecords), newRecord);
        newRecord.setIndexPosition(currentNumRecords);
        writeNumRecordsHeader(currentNumRecords + 1);

        synchronized (memIndex) {
            memIndex.put(key, newRecord);
        }
    }

    /**
     * Removes the record from the index. Replaces the target with the entry at the
     * end of the index.
     */
    protected void deleteEntryFromIndex(String key, RecordHeader header, int currentNumRecords)
            throws IOException, RecordsFileException {
        if (header.getIndexPosition() != (currentNumRecords - 1)) {
            String lastKey = readKeyFromIndex(currentNumRecords - 1);
            RecordHeader last = keyToRecordHeader(lastKey);
            last.setIndexPosition(header.getIndexPosition());
            writeUTF(indexPositionToKeyFp(last.getIndexPosition()), lastKey);
            writeRecordHeader(indexPositionToRecordHeaderFp(last.getIndexPosition()), last);
        }
        writeNumRecordsHeader(currentNumRecords - 1);
        synchronized (memIndex) {
            memIndex.remove(key);
        }
    }

    public Collection getKeysStartingWith(String filter) {
        Collection result = new ArrayList();
        synchronized (memIndex) {
            Iterator keys = memIndex.keySet().iterator();
            while (keys.hasNext()) {
                String key = (String) keys.next();
                if (key.startsWith(filter)) {
                    result.add(key);
                }
            }
        }
        return result;
    }


    // Checks to see if there is space for and additional index entry. If
    // not, space is created by moving records to the end of the file.
    protected void insureIndexSpace(int requiredNumRecords)
            throws RecordsFileException, IOException {
        int originalFirstDataCapacity;
        int currentNumRecords = getNumRecords();
        long endIndexPtr = indexPositionToKeyFp(requiredNumRecords);

        if ((endIndexPtr > getFileLength()) && (currentNumRecords == 0)) {
            setFileLength(endIndexPtr);
            dataStartPtr = endIndexPtr;
            writeDataStartPtrHeader(dataStartPtr);

            return;
        }

        while (endIndexPtr > dataStartPtr) {
            RecordHeader first = getRecordAt(dataStartPtr);
            byte[] data = readRecordData(first);
            first.setDataPointer(getFileLength());

            // If first.dataCapacity is set to the actual data count BEFORE resetting dataStartPtr,
            // and there is free space in 'first', then dataStartPtr will not be reset to the start
            // of the second record.  Capture the capacity first and use it to perform the reset.
            originalFirstDataCapacity = first.getDataCapacity();
            first.setDataCapacity(data.length);
            setFileLength(first.getDataPointer() + data.length);
            writeRecordData(first, data);
            writeRecordHeaderToIndex(first);
            dataStartPtr += originalFirstDataCapacity;
            writeDataStartPtrHeader(dataStartPtr);
        }
    }

    /**
     * Returns a file pointer in the index pointing to the first byte
     * in the key located at the given index position.
     */
    protected long indexPositionToKeyFp(int pos) {
        return FILE_HEADERS_REGION_LENGTH + (indexEntryLength * pos);
    }

    protected long readDataStartHeader() throws IOException {
        file.seek(DATA_START_HEADER_LOCATION);
        return file.readLong();
    }

    protected int readNumRecordsHeader() throws IOException {
        file.seek(NUM_RECORDS_HEADER_LOCATION);
        return file.readInt();
    }

    protected byte[] readRecordData(String key) throws IOException, RecordsFileException {
        return readRecordData(keyToRecordHeader(key));
    }

    protected byte[] readRecordData(RecordHeader header) throws IOException {
        byte[] buf = new byte[header.getDataCount()];
        file.seek(header.getDataPointer());
        file.readFully(buf);
        return buf;
    }

    protected void writeDataStartPtrHeader(long dataStartPtr) throws IOException {
        writeLong(DATA_START_HEADER_LOCATION, dataStartPtr);
    }

    protected void writeNumRecordsHeader(int numRecords) throws IOException {
        writeInt(NUM_RECORDS_HEADER_LOCATION, numRecords);
    }

    /**
     * Updates the contents of the given record. A RecordsFileException is thrown if the new data does not
     * fit in the space allocated to the record. The header's data count is updated, but not
     * written to the file.
     */
    protected void writeRecordData(RecordHeader header, RecordWriter rw) throws IOException, RecordsFileException {
        if (rw.getDataLength() > header.getDataCapacity()) {
            throw new RecordsFileException("Record data does not fit");
        }
        header.setDataCount(rw.getDataLength());
        writeRecordWriter(header, rw);
    }

    /**
     * Updates the contents of the given record. A RecordsFileException is thrown if the new data does not
     * fit in the space allocated to the record. The header's data count is updated, but not
     * written to the file.
     */
    protected void writeRecordData(RecordHeader header, byte[] data)
            throws IOException, RecordsFileException {
        if (data.length > header.getDataCapacity()) {
            throw new RecordsFileException("Record data does not fit");
        }
        header.setDataCount(data.length);
        writeByteArray(header.getDataPointer(), data);
    }

    protected void writeRecordHeaderToIndex(RecordHeader header) throws IOException {
        writeRecordHeader(indexPositionToRecordHeaderFp(header.getIndexPosition()), header);
        synchronized (memIndex) {
            memIndex.put(readKeyFromIndex(header.getIndexPosition()), header);
        }
    }

    /**
     * Returns a file pointer in the index pointing to the first byte
     * in the record pointer located at the given index position.
     */
    long indexPositionToRecordHeaderFp(int pos) {
        return indexPositionToKeyFp(pos) + maxKeyLength;
    }

    /**
     * Reads the ith key from the index.
     */
    protected String readKeyFromIndex(int position) throws IOException {
        file.seek(indexPositionToKeyFp(position));
        return file.readUTF();
    }

    /**
     * Reads the ith record header from the index.
     */
    protected RecordHeader readRecordHeaderFromIndex(int position) throws IOException {
        file.seek(indexPositionToRecordHeaderFp(position));
        return RecordHeader.readHeader(file);
    }


    public synchronized void reorganize() throws IOException, RecordsFileException {
        long length = file.length();
        long moveAddress = dataStartPtr;
        long nextTargetAddress = dataStartPtr;
        while (length > nextTargetAddress) {
            RecordHeader header = getRecordAt(nextTargetAddress);
            nextTargetAddress = header.getDataPointer() + header.getDataCapacity();
            moveAddress = moveDataBack(header, moveAddress);
        }
        setFileLength(moveAddress);
    }

    // return next address to move data
    private synchronized long moveDataBack(RecordHeader header, long moveAddress) throws IOException, RecordsFileException {
        if (header.getDataPointer() > moveAddress) {
            byte[] buf = readRecordData(header);
            long maxAddress = header.getDataPointer() + header.getDataCapacity();
            int size = (new Double(buf.length * additionalSpace)).intValue();

            if ((moveAddress + size) > maxAddress) {
                size = (int) (maxAddress - moveAddress);
            }

            RecordHeader newHeader = new RecordHeader(moveAddress, size);
            writeRecordData(newHeader, buf);
            newHeader.setIndexPosition(header.getIndexPosition());
            writeRecordHeaderToIndex(newHeader);
            return newHeader.getDataPointer() + newHeader.getDataCapacity();

            // no move, just resize
        } else if (header.getDataPointer() == moveAddress) {
            int size = (new Double(header.getDataCount() * additionalSpace)).intValue();
            if (size > header.getDataCapacity()) {
                size = header.getDataCapacity();
            }
            header.setDataCapacity(size);
            writeRecordHeaderToIndex(header);
            return (header.getDataPointer() + header.getDataCapacity());
        } else {
            throw new RecordsFileException("moveDataBack address error");
        }
    }

    public void setTrace(Trace trace) {
        this.trace = trace;
    }


    /**
     * *******************************************  LOGGING  ************************************************
     */

    protected void writeByteArray(long position, byte[] datas) throws IOException {
        file.seek(position);
        file.write(datas);
    }

    protected void writeOutputStream(long position, DbByteArrayOutputStream outputStream) throws IOException {
        file.seek(position);
        outputStream.writeTo(file);
    }

    protected void writeRecordWriter(RecordHeader header, RecordWriter recordWriter) throws IOException {
        file.seek(header.getDataPointer());
        recordWriter.writeTo(file);
    }

    protected void writeRecordHeader(long position, RecordHeader recordHeader) throws IOException {
        file.seek(position);
        recordHeader.write(file);
    }

    protected void writeInt(long position, int i) throws IOException {
        file.seek(position);
        file.writeInt(i);
    }

    protected void writeLong(long position, long l) throws IOException {
        file.seek(position);
        file.writeLong(l);
    }

    protected void writeUTF(long position, String s) throws IOException {
        file.seek(position);
        file.writeUTF(s);
    }


    /**
     * ******************************   ABSTRACT   *****************************************
     */


    public synchronized int getNumRecords() {
        return memIndex.size();
    }

    public synchronized Iterator enumerateKeys() {
        return memIndex.keySet().iterator();
    }

    public synchronized boolean recordExists(String key) {
        return memIndex.containsKey(key);
    }

    /**
     * Returns the record to which the target file pointer belongs - meaning the specified location
     * in the file is part of the record data of the RecordHeader which is returned.  Returns null if
     * the location is not part of a record. (O(n) mem accesses)
     */
    protected RecordHeader getRecordAt(long targetFp) {
        Iterator elements = memIndex.values().iterator();
        while (elements.hasNext()) {
            RecordHeader next = (RecordHeader) elements.next();
            if ((targetFp >= next.getDataPointer()) &&
                (targetFp < (next.getDataPointer() +
                             (long) next.getDataCapacity()))) {
                return next;
            }
        }
        return null;
    }

    /**
     * This method searches the file for free space and then returns a RecordHeader
     * which uses the space. (O(n) memory accesses)
     */
    protected RecordHeader allocateRecord(String key, int dataLength)
            throws RecordsFileException, IOException {
        // search for empty space
        RecordHeader newRecord = null;
        Iterator elements = memIndex.values().iterator();
        while (elements.hasNext()) {
            RecordHeader next = (RecordHeader) elements.next();
            if (dataLength <= next.getFreeSpace()) {
                newRecord = next.split();
                writeRecordHeaderToIndex(next);
                break;
            }
        }
        if (newRecord == null) {
            // append record to end of file - grows file to allocate space
            long fp = getFileLength();
            setFileLength(fp + dataLength);
            newRecord = new RecordHeader(fp, dataLength);
        }
        return newRecord;
    }

    /**
     * Maps a key to a record header by looking it up in the in-memory index.
     */
    protected RecordHeader keyToRecordHeader(String key) throws RecordsFileException {
        RecordHeader h = (RecordHeader) memIndex.get(key);
        if (h == null) {
            throw new RecordsFileException("Key not found: " + key);
        }
        return h;
    }

    protected void readHeadersInMemIndex() throws IOException {
        int numRecords = readNumRecordsHeader();
        memIndex = new HashMap(numRecords);

        synchronized (memIndex) {
            for (int i = 0; i < numRecords; i++) {
                String key = readKeyFromIndex(i);
                RecordHeader header = readRecordHeaderFromIndex(i);
                header.setIndexPosition(i);
                memIndex.put(key, header);
            }
        }
    }

    /**
     * ***************************************************************************************************
     */

    public boolean equals(Object object) {
        RecordsFile candidate;
        try {
            candidate = (RecordsFile) object;
        } catch (ClassCastException cce) {
            System.out.println("RecordsFile equals : candidate is not a RecordsFile instance");
            return false;
        }

        if (memIndex.size() != candidate.memIndex.size()) {
            System.out.println("RecordsFile equals : not the same number of key");
            return false;
        }

        Iterator keys = memIndex.keySet().iterator();
        while (keys.hasNext()) {
            String key = (String) keys.next();
            if (!candidate.memIndex.containsKey(key)) {
                System.out.println("RecordsFile equals : one contains a key, other don't");
                return false;
            }
            RecordHeader rh = (RecordHeader) memIndex.get(key);
            RecordHeader crh = (RecordHeader) candidate.memIndex.get(key);
            if (!rh.equals(crh)) {
                System.out.println("RecordsFile equals : recordHeader not equal for key " + key);
                System.out.println(rh.toString());
                System.out.println(crh.toString());
                return false;
            }
            try {
                if (!JalistoUtils.areEquals(readRecordData(rh), candidate.readRecordData(crh))) {
                    System.out.println("RecordsFile equals : datas not equals for key " + key);
                    return false;
                }
            } catch (IOException ioe) {
                ioe.printStackTrace();
                return false;
            }
        }

        return true;
    }
}
